/*****************************************************************************
 * C/C++ Headers
 *****************************************************************************/

#include <stdio.h>
#include <ctype.h>

/*****************************************************************************
 * Local Headers
 *****************************************************************************/

#include "cfg.h"
#include "regex.h"
#include "util.h"

/*****************************************************************************
 * Local Defines
 *****************************************************************************/

#define SECTION_HEADER_REGEX \
   "^\\s*\\[\\s*([a-z_]+)\\s*\\]\\s*$"

#define SECTION_KV_REGEX \
   "^\\s*([a-z_][a-z_0-9]*)\\s*=\\s*(.*)$"

#define SECTION_COMMENT_REGEX \
   "^\\s*#"

#define SECTION_BLANK_REGEX \
   "^\\s*$"

/*****************************************************************************
 * Local Functions
 *****************************************************************************/

static int getSection(
   EntryVector ev,
   CfgSections &sections,
   int &index,
   string &errMsg
)
{
   Entry e = ev.at(0);

   string sectionName = Util::toLowerCase(e.mStrings.at(1));

   CfgSection theSection(sectionName);

   sections.push_back(theSection);

   index = sections.size() - 1;

   return(1);
}

static int getKV(
   EntryVector ev,
   CfgSections &sections,
   int &index,
   string &errMsg
)
{
   Entry e = ev.at(0);

   string key   = Util::toLowerCase(e.mStrings.at(1));
   string value = Util::trim(e.mStrings.at(2));

   if(index < 0)
   {
      errMsg = "No section preceding key: [";
      errMsg += key;
      errMsg += "], value: [";
      errMsg += value;
      errMsg += "]";

      return(0);
   }

   if(value == "")
   {
      errMsg = "No value supplied for key: [" + key + "]";
      return(0);
   }

   sections.at(index).add(key, value);

   return(1);
}

static int getComment(
   EntryVector ev,
   CfgSections &sections,
   int &index,
   string &errMsg
)
{
   return(1);
}

static int getBlank(
   EntryVector ev,
   CfgSections &sections,
   int &index,
   string &errMsg
)
{
   return(1);
}

/*****************************************************************************
 * Classes: CfgLine, CfgKeyValue, CfgSection
 *****************************************************************************/

// -- CfgLine --

CfgLine::CfgLine(int lineNumber, string line) :
   mLineNumber(lineNumber),
   mLine(line)
{
}

void CfgLine::dump()
{
   printf(
      "%3d) '%s'\n",
      mLineNumber,
      Util::rtrim(mLine).c_str()
   );
}

// -- CfgLines --

void CfgLines::add(CfgLine theLine)
{
   mLines.push_back(theLine);
}

void CfgLines::dump()
{
   int n = mLines.size();

   printf("------------------------------------------------\n");
   printf("CfgLines - Dump\n");
   printf("------------------------------------------------\n");
   printf("\n");

   for(int i = 0; i < n; i ++)
      mLines.at(i).dump();

   printf("\n");
}

// -- CfgKeyValue --

CfgKeyValue::CfgKeyValue(string key, string value) :
   mKey(key), mValue(value)
{
}

void CfgKeyValue::dump()
{
   printf("   %-15s = '%s'\n", mKey.c_str(), mValue.c_str());
}

CfgSection::CfgSection(string sectionName) : mSectionName(sectionName)
{
}

void CfgSection::add(string key, string value)
{
   CfgKeyValue keyAndValue(key, value);

   mKV.push_back(keyAndValue);
}

void CfgSection::dump()
{
   printf("sectionName: [%s]\n", mSectionName.c_str());

   int n = mKV.size();

   for(int i = 0; i < n; i ++)
      mKV.at(i).dump();

   printf("\n");
}

/*****************************************************************************
 * Definition, Class: 'Cfg'
 *****************************************************************************/

int Cfg::getLines(
   const char *fileName,
   CfgLines    &linesObject,
   string      &errMsg
)
{
   errMsg  = "";
   FILE *f = NULL;

   if((f = fopen(fileName, "r")) == NULL)
   {
      errMsg =
         errMsg +
         "Could not open file: [" +
         fileName                 +
         "]";

      return(0);
   }

   int go           = 1;
   int lineNumber   = 1;
   string theString = "";

   while(go)
   {
      int c         = fgetc(f);
      int doAddLine = 0;

      if(c == EOF)
      {
         if(theString.length() > 0)
         {
            doAddLine = 1;
         }

         go = 0;
      }
      else
      {
         char ch    = (char)c;
         theString += ch;

         if(ch == '\n')
         {
            doAddLine = 1;
         }
      }

      if(doAddLine)
      {
         CfgLine cfgLine(lineNumber++, theString);

         linesObject.add(cfgLine);

         theString = "";
      }
   }

   fclose(f);

   if(0)
   {
      linesObject.dump();
   }

   return(1);
}


int Cfg::parse(
   CfgLines    &linesObject,
   CfgSections &sections,
   string      &errMsg
)
{
   struct TableEntry
   {
      const char *regex;
      int nEntries;
      int (*process)(EntryVector, CfgSections &, int &, string &);
   };

   // ----------------------------------------------------
   // There are 4 types of lines we'll tolerate:
   //
   //    - A section header
   //
   //      [settings]
   //
   //    - A key/value pair (within a section)
   //
   //      timeout_value = 15
   //
   //    - A lines whose first non-whitespace character
   //      is a '#' (a comment)
   //
   //      # log_file = c:\documents and settings\log.txt
   //
   //    - A line composed entirely of whitespace
   //
   // Below, we set up a mapping between each
   // regex and the handler function we call when we
   // encounter said regex.
   // ----------------------------------------------------

   TableEntry entries[] =
   {
      {
         SECTION_HEADER_REGEX,
         2,
         getSection
      },
      {
         SECTION_KV_REGEX,
         3,
         getKV
      },
      {
         SECTION_COMMENT_REGEX,
         1,
         getComment
      },
      {
         SECTION_BLANK_REGEX,
         1,
         getBlank
      }
   };

   int nEntries  = USZ(entries);
   int n         = linesObject.mLines.size();
   int index     = -1;

   // For each line, we try to identify the line
   // as being one of the four types mentioned
   // above.

   for(int i = 0; i < n; i ++)
   {
      CfgLine cfgLine = linesObject.mLines.at(i);

      const char *line = cfgLine.mLine.c_str();
      int lineNumber   = cfgLine.mLineNumber;
      int found        = 0;

      for(int j = 0; j < nEntries; j ++)
      {
         EntryVector ev;

         int mValue = Regex::match(
            entries[j].regex,
            1,
            line,
            ev,
            errMsg
         );

         if(!mValue)
         {
            // errMsg will be empty if this was just a
            // non-match of the regex as opposed to
            // a real error (i.e. an invalid pattern)

            if(errMsg != "")
               return(0);
         }
         else
         {
            // We found a regex that matched our line

            int size = ev.size();

            // This should never happen as all our patterns
            // are anchored but one never knows ...

            if(size != 1)
            {
               errMsg  = Util::toString(
                  "Line: [%d], multiple matches for: [",
                  lineNumber
               )                 +
               entries[j].regex  +
               "] in: ["         +
               Util::rtrim(line) +
               "]";

               return(0);
            }

            Entry e = ev.at(0);
            size    = e.mStrings.size();

            // See if the number of captured strings
            // is what we expect for this spec entry

            if(size != entries[j].nEntries)
            {
               errMsg =
                  Util::toString(
                     "Line: [%d], ", lineNumber
                  )                                          +
                  Util::toString(
                     "expected: [%d] matches for: [",
                     entries[j].nEntries
                  )                                          +
                  entries[j].regex                           +
                  "] in: ["                                  +
                  Util::rtrim(line)                          +
                  "], "                                      +
                  Util::toString("not: [%d]", size);

               return(0);
            }

            found = 1;

            // Let the spec function do whatever it wants
            // to 'sections' ...

            if(!entries[j].process(ev, sections, index, errMsg))
               return(0);
         }
      }

      if(!found)
      {
         // Not a type of line we recognize

         errMsg =
            Util::toString(
               "Line: [%d], invalid line: [", lineNumber
            )    +
            Util::rtrim(line) +
            "]";

         return(0);
      }
   }

   return(1);
}

int Cfg::checkFrequency(
   CfgItemFrequency frequency,
   int              counted,
   string           title,
   string           target,
   string           &errMsg
)
{
   errMsg            = "";
   string errorRange = "";

   switch(frequency)
   {
      case CFG_ZERO_OR_ONE:
      {
         if(counted > 1)
         {
            errorRange = "zero or one instances";
         }
      }
      break;

      case CFG_ZERO_OR_MORE:
      {
      }
      break;

      case CFG_EXACTLY_ONE:
      {
         if(counted != 1)
            errorRange = "one instance";
      }
      break;

      case CFG_ONE_OR_MORE:
      {
         if(counted < 1)
            errorRange = "one or more instances";
      }
      break;

      default:
      {
         errMsg = Util::toString(
            "invalid frequency: [%d]", frequency
         );

         return(0);
      }
      break;
   }

   if(errorRange == "")
      return(1);

   errMsg +=
      "Expected " +
      errorRange +
      " of " +
      title + ": [" +
      target +
      "], " +
      Util::toString("found: [%d]", counted);

   return(0);
}

string Cfg::freqToString(CfgItemFrequency freq)
{
   if(freq == CFG_ZERO_OR_ONE)
      return("0/1");

   if(freq == CFG_ZERO_OR_MORE)
      return("0+");

   if(freq == CFG_EXACTLY_ONE)
      return("1");

   if(freq == CFG_ONE_OR_MORE)
      return("1+");

   return("Unknown");
}

int Cfg::isValidSection(string sectionName, int nSpec, CfgItemSpec *spec)
{
   // Look for 'sectionName' among all sections
   // specified in 'spec'.

   for(int i = 0; i < nSpec; i ++)
   {
      if(Util::imatch(sectionName, spec[i].sectionName))
      {
         return(1);
      }
   }

   return(0);
}

int Cfg::isValidKey(
   string sectionName, string key, int nSpec, CfgItemSpec *spec
)
{
   for(int i = 0; i < nSpec; i ++)
   {
      if(Util::imatch(sectionName, spec[i].sectionName) && spec[i].keyName)
      {
         if(Util::imatch(key, spec[i].keyName))
         {
            return(1);
         }
      }
   }

   return(0);
}


int Cfg::doString(string key, string value, string &sValue, string &errMsg)
{
   sValue = value;
   return(1);
}

int Cfg::doBoolean(
   string key,
   string value,
   string &sValue,
   string &errMsg
)
{
   int returnValue = 0;

   switch(value[0])
   {
      case '1': case 't': case 'T': case 'y': case 'Y':
      {
         sValue      = CFG_YES;
         returnValue = 1;
      }
      break;

      case '0': case 'f': case 'F': case 'n': case 'N':
      {
         sValue      = CFG_NO;
         returnValue = 1;
      }
      break;

      default:
      {
         errMsg  = "Illegal value: [";
         errMsg += value + "] for boolean: [" + key + "]";
      }
      break;
   }

   return(returnValue);
}

int Cfg::isTrue(string value)
{
   return(value == CFG_YES);
}

int Cfg::isFalse(string value)
{
   return(value == CFG_NO);
}

int Cfg::applyDefaults(
   void *voidConfig,
   CfgItemSpec *spec,
   int nSpec,
   string &errMsg
)
{
   int i = 0;

   for(i = 0; i < nSpec; i ++)
   {
      if(spec[i].keyName && spec[i].itemDefaultValue && spec[i].cbf)
      {
         int status = spec[i].cbf(
            voidConfig,
            spec[i].keyName,
            spec[i].itemDefaultValue,
            errMsg
         );

         if(!status)
            return(0);
      }
   }

   return(1);
}

int Cfg::processConfig(
   void *voidConfig,
   int doApplyDefaults,
   CfgSections sections,
   CfgItemSpec spec[],
   int nSpec,
   string &errMsg
)
{
   if(doApplyDefaults)
   {
      if(!Cfg::applyDefaults(voidConfig, spec, nSpec, errMsg))
         return(0);
   }

   int nSections = sections.size();
   int i         = 0;
   int j         = 0;
   int k         = 0;

   // ----------------------------------------------
   // Look through 'sections' at the set of
   // section names and key names.  Ensure that
   // they are all ones we recognize.
   // ----------------------------------------------

   for(i = 0; i < nSections; i ++)
   {
      CfgSection theSection = sections.at(i);
      string sectionName    = theSection.mSectionName;

      if(!Cfg::isValidSection(sectionName, nSpec, spec))
      {
         errMsg  = "Not a valid section name: [";
         errMsg += sectionName + "]";

         return(0);
      }

      // sectionName must be a recognized section name.
      // Now, make sure all keys within this section
      // are ones we recognize

      int nKV = theSection.mKV.size();

      for(j = 0; j < nKV; j ++)
      {
         string theKey = theSection.mKV.at(j).mKey;

         if(!Cfg::isValidKey(sectionName, theKey, nSpec, spec))
         {
            errMsg  = "Section: [" + sectionName;
            errMsg += "], invalid key: [";
            errMsg += theKey + "]";

            return(0);
         }
      }
   }

   // Now, we know for sure that all sections and keys
   // in 'sections' are good.  We now look to see if
   // the counts are right - first for our sections,
   // then for our keys within a section.

   for(i = 0; i < nSpec; i ++)
   {
      CfgItemSpec theItem = spec[i];
      string sectionName  = theItem.sectionName;

      if(!theItem.keyName)
      {
         // A section, count the number of
         // supplied sections that match

         int nMatch = 0;

         for(j = 0; j < nSections; j ++)
         {
            if(Util::imatch(sectionName, sections.at(j).mSectionName))
            {
               nMatch++;
            }
         }

         int status = Cfg::checkFrequency(
            theItem.frequency, nMatch, "section", sectionName, errMsg
         );

         if(!status)
            return(0);
      }
      else
      {
         string keyName = theItem.keyName;

         for(j = 0; j < nSections; j ++)
         {
            if(Util::imatch(sectionName, sections.at(j).mSectionName))
            {
               int nKV    = sections.at(j).mKV.size();
               int nMatch = 0;

               for(k = 0; k < nKV; k ++)
               {
                  string theKey = sections.at(j).mKV.at(k).mKey;

                  if(Util::imatch(keyName, theKey))
                  {
                     nMatch++;
                  }
               }

               int status = Cfg::checkFrequency(
                  theItem.frequency,
                  nMatch,
                  "(section: '" + sectionName + "') key",
                  keyName,
                  errMsg
               );

               if(!status)
                  return(0);
            }
         }
      }
   }

   // Ok, we found the right sections and keys and the
   // right numbers of each.  We now validate the values
   // supplied and apply them.

   for(i = 0; i < nSections; i ++)
   {
      CfgSection theSection = sections.at(i);
      string sectionName    = theSection.mSectionName;

      for(j = 0; j < nSpec; j ++)
      {
         if(!spec[j].keyName)
         {
            // This is a section spec, the one we're
            // looking for?

            if(Util::imatch(sectionName, spec[j].sectionName))
            {
               // If there's a handler function
               // we call it.  This allows for an
               // arbitrary action to be performed
               // after the section name we're working
               // on is known but before we process
               // the section's key/value pairs.

               if(spec[j].cbf)
               {
                  int status = spec[j].cbf(
                     voidConfig,
                     sectionName,
                     "",
                     errMsg
                  );
                     
                  if(!status)
                     return(0);
               }

               break;
            }
         }
      }

      // Now for all our key/value pairs

      int nKV = sections.at(i).mKV.size();

      for(j = 0; j < nKV; j ++)
      {
         string key   = sections.at(i).mKV.at(j).mKey;
         string value = sections.at(i).mKV.at(j).mValue;

         for(k = 0; k < nSpec; k ++)
         {
            if(spec[k].keyName)
            {
               if(Util::imatch(sectionName, spec[k].sectionName) &&
                  Util::imatch(key, spec[k].keyName))
               {
                  if(spec[k].cbf)
                  {
                     int status = spec[k].cbf(
                        voidConfig,
                        key,
                        value,
                        errMsg
                     );

                     if(!status)
                        return(0);
                  }

                  break;
               }
            }
         }
      }
   }

   return(1);
}
