#ifndef _CFG_H_
#   define _CFG_H_

#include <string>
#include <vector>

using std::string;
using std::vector;

// -----------------------------------
// Defines and Enums
// -----------------------------------

#define CFG_YES "y"
#define CFG_NO  "n"

// -----------------------------------
// Frequency specifications
// -----------------------------------

typedef enum
{
   CFG_ZERO_OR_ONE,
   CFG_ZERO_OR_MORE,
   CFG_EXACTLY_ONE,
   CFG_ONE_OR_MORE

} CfgItemFrequency;

// -----------------------------------
// A line along with its line number
// -----------------------------------

struct CfgLine
{
   int mLineNumber;
   string mLine;

   CfgLine(int lineNumber, string line);
   void dump();
};

// -----------------------------------
// A set of CfgLine objects
// -----------------------------------

struct CfgLines
{
   vector<CfgLine> mLines;

   void add(CfgLine theLine);
   void dump();
};

// -----------------------------------
// A key/value pair
// -----------------------------------

struct CfgKeyValue
{
   string mKey;
   string mValue;

   CfgKeyValue(string key, string value);
   void dump();
};

// -----------------------------------
// A group (section) name and a set of
// associated key/value pairs
// -----------------------------------

struct CfgSection
{
   string mSectionName;
   vector<CfgKeyValue> mKV;

   CfgSection(string sectionName);
   void add(string key, string value);

   void dump();
};

// -----------------------------------
// An array of CfgSection objects
// -----------------------------------

typedef vector<CfgSection> CfgSections;

/* -------------------------------------------------------------

   A config file element

   For example:
  
      { "verbose", CFG_YES, CFG_ZERO_OR_ONE, verbose }
                                        ^
   The function pointer ----------------+

   Can either be a static method, as in ...

      class SomeClass
      {
         public:
            int verbose(
               string value, ConfigBase &cfg, string errMsg
            );
      };

      int SomeClass::verbose(
         string value, ConfigBase &cfg, string errMsg
      )
      {
         // ...
      }

   Or a stand-alone function, as in ...

      static int verbose
         string value, ConfigBase &cfg, string errMsg
      )
      {
         // ...
      }

   The handler function's job is to either:

      - apply 'value' to our config struct ... or ...
      - report an error

   ------------------------------------------------------------- */

typedef int (*CallBackFunction)(void *, string, string, string &);

struct CfgItemSpec
{
   const char *sectionName;
   const char *keyName;

   CfgItemFrequency frequency;

   const char *itemDefaultValue;

   CallBackFunction cbf;
};

// ----------------------------------------------------
// Definition of our config utilities class
// ----------------------------------------------------

class Cfg
{
   public:
      static int getLines(
         const char *fileName,
         CfgLines    &linesObject,
         string      &errMsg
      );

      static int parse(
         CfgLines    &linesObject,
         CfgSections &sections,
         string      &errMsg
      );

      static int checkFrequency(
         CfgItemFrequency frequency,
         int              counted,
         string           title,
         string           target,
         string           &errMsg
      );

      static string freqToString(CfgItemFrequency freq);

      static int isValidSection(
         string sectionName, int nSpec, CfgItemSpec *spec
      );

      static int isValidKey(
         string sectionName, string key, int nSpec, CfgItemSpec *spec
      );

      static int doString(
         string key, string value, string &sValue, string &errMsg
      );

      static int doBoolean(
         string key,
         string value,
         string &sValue,
         string &errMsg
      );

      static int isTrue(string value);
      static int isFalse(string value);

      static int applyDefaults(
         void        *voidConfig,
         CfgItemSpec *spec,
         int         nSpec,
         string      &errMsg
      );

      static int processConfig(
         void        *voidConfig,
         int         doApplyDefaults,
         CfgSections sections,
         CfgItemSpec spec[],
         int         nSpec,
         string      &errMsg
      );
};

#endif
