/*****************************************************************************
 * C/C++ Headers
 *****************************************************************************/

#include <stdio.h>

/*****************************************************************************
 * Local Headers
 *****************************************************************************/

#include "cfg.h"
#include "util.h"
#include "config.h"

#define CONFIG_FIELD_SIZE (12)

/*****************************************************************************
 * Class: Config
 *****************************************************************************/

Config::Config()
{
   mCurrentEntry = &mDefaultEntry;
}

/*****************************************************************************
 * Utility Functions
 *****************************************************************************/

string Config::entryToString(Entry &theEntry, string title)
{
   string s = "[entry";

   if(title != "")
      s += " " + title;

   s += "]\n";

   s += kvToString(CONFIG_TO,      theEntry.mTo)      + "\n";
   s += kvToString(CONFIG_UPDATE,  theEntry.mUpdate)  + "\n";
   s += kvToString(CONFIG_FROM,    theEntry.mFrom)    + "\n";
   s += kvToString(CONFIG_RECURSE, theEntry.mRecurse) + "\n";
   s += kvToString(CONFIG_HIDDEN,  theEntry.mHidden)  + "\n";
   s += kvToString(CONFIG_SYSTEM,  theEntry.mSystem)  + "\n";

   int nExclude = theEntry.mExcludeList.size();
   int i        = 0;

   if(nExclude == 0)
   {
      s += kvToString(CONFIG_EXCLUDE, "") + "\n";
   }
   else
   {
      s +=
         "   " +
         Util::rpad(CONFIG_EXCLUDE, CONFIG_FIELD_SIZE) + " = {\n";

      for(i = 0; i < nExclude; i ++)
      {
         s += "   " + 
            Util::toString("   %03d)", i) + " '" +
            theEntry.mExcludeList.at(i) + "'\n";
      }

      s += "   }\n";
   }

   return(s);
}

void Config::dumpEntry(Entry &theEntry, string title)
{
   printf("%s\n", entryToString(theEntry, title).c_str());
}

string Config::toString(string title, int showDefaultEntry)
{
   string s = "";

   s += "------------------------------------------------\n";
   s += "Config - " + title + " - Dump\n";
   s += "------------------------------------------------\n";
   s += "\n";

   s += "[options]\n";

   s += kvToString(CONFIG_TO,           mOptions.mTo)          + "\n";
   s += kvToString(CONFIG_UPDATE,       mOptions.mUpdate)      + "\n";
   s += kvToString(CONFIG_HIDDEN,       mOptions.mHidden)      + "\n";
   s += kvToString(CONFIG_SYSTEM,       mOptions.mSystem)      + "\n";
   s += kvToString(CONFIG_VERBOSE,      mOptions.mVerbose)     + "\n";
   s += kvToString(CONFIG_FORCE,        mOptions.mForce)       + "\n";
   s += kvToString(CONFIG_LOG,          mOptions.mLog)         + "\n";
   s += kvToString(CONFIG_CONSOLE,      mOptions.mConsole)     + "\n";
   s += kvToString(CONFIG_SHOW_ERROR,   mOptions.mShowError)   + "\n";
   s += kvToString(CONFIG_SHOW_SUCCESS, mOptions.mShowSuccess) + "\n";

   if(showDefaultEntry)
   {
      s += "\n";
      s += entryToString(mDefaultEntry, "(default)");
   }

   int size = mEntries.size();

   for(int i = 0; i < size; i ++)
   {
      s += "\n";
      s += entryToString(mEntries.at(i), "");
   }

   return(s);
}

void Config::dump(string title)
{
   printf("%s\n", toString(title, 1).c_str());
}

string Config::kvToString(string key, string value)
{
   string s =
      "   " + Util::rpad(key, CONFIG_FIELD_SIZE);

   s += " = " + (value == "" ? "[Not Set]" : "'" + value + "'");

   return(s);
}

void Config::debug(string msg)
{
   if(0)
      printf("[DEBUG] %s\n", msg.c_str());
}

void Config::debug(string section, string key, string value)
{
   string s =
      "Section: [" + section + "], " +
      "Key: [" + key + "], Value: [" + value + "]";

   Config::debug(s);
}

/*****************************************************************************
 * Callback Functions
 *****************************************************************************/

int Config::optionsBegin(
   void *voidPtr, string key, string value, string &errMsg
)
{
   debug("Begin: 'options'");

   return(1);
}

int Config::optionsTo(
   void *voidPtr, string key, string value, string &errMsg
)
{
   debug(CONFIG_OPTIONS, key, value);

   Config *c = (Config *)voidPtr;

   return(Cfg::doString(key, value, c->mOptions.mTo, errMsg));
}

int Config::optionsUpdate(
   void *voidPtr, string key, string value, string &errMsg
)
{
   debug(CONFIG_OPTIONS, key, value);

   Config *c = (Config *)voidPtr;

   return(Cfg::doBoolean(key, value, c->mOptions.mUpdate, errMsg));
}

int Config::optionsHidden(
   void *voidPtr, string key, string value, string &errMsg
)
{
   debug(CONFIG_OPTIONS, key, value);

   Config *c = (Config *)voidPtr;

   return(Cfg::doBoolean(key, value, c->mOptions.mHidden, errMsg));
}

int Config::optionsSystem(
   void *voidPtr, string key, string value, string &errMsg
)
{
   debug(CONFIG_OPTIONS, key, value);

   Config *c = (Config *)voidPtr;

   return(Cfg::doBoolean(key, value, c->mOptions.mSystem, errMsg));
}

int Config::optionsVerbose(
   void *voidPtr, string key, string value, string &errMsg
)
{
   debug(CONFIG_OPTIONS, key, value);

   Config *c = (Config *)voidPtr;

   return(Cfg::doBoolean(key, value, c->mOptions.mVerbose, errMsg));
}

int Config::optionsForce(
   void *voidPtr, string key, string value, string &errMsg
)
{
   debug(CONFIG_OPTIONS, key, value);

   Config *c = (Config *)voidPtr;

   return(Cfg::doBoolean(key, value, c->mOptions.mForce, errMsg));
}


int Config::optionsLog(
   void *voidPtr, string key, string value, string &errMsg
)
{
   debug(CONFIG_OPTIONS, key, value);

   Config *c = (Config *)voidPtr;

   return(Cfg::doBoolean(key, value, c->mOptions.mLog, errMsg));
}

int Config::optionsConsole(
   void *voidPtr, string key, string value, string &errMsg
)
{
   debug(CONFIG_OPTIONS, key, value);

   Config *c = (Config *)voidPtr;

   return(Cfg::doBoolean(key, value, c->mOptions.mConsole, errMsg));
}

int Config::optionsShowError(
   void *voidPtr, string key, string value, string &errMsg
)
{
   debug(CONFIG_OPTIONS, key, value);

   Config *c = (Config *)voidPtr;

   return(Cfg::doBoolean(key, value, c->mOptions.mShowError, errMsg));
}

int Config::optionsShowSuccess(
   void *voidPtr, string key, string value, string &errMsg
)
{
   debug(CONFIG_OPTIONS, key, value);

   Config *c = (Config *)voidPtr;

   return(Cfg::doBoolean(key, value, c->mOptions.mShowSuccess, errMsg));
}

int Config::entryBegin(
   void *voidPtr, string key, string value, string &errMsg
)
{
   debug("Begin: 'entry'");

   Config *c              = (Config *)voidPtr;
   Config::Entry theEntry = c->mDefaultEntry;

   c->mEntries.push_back(theEntry);

   c->mCurrentEntry = &(c->mEntries.at(c->mEntries.size() - 1));

   return(1);
}

int Config::entryTo(
   void *voidPtr, string key, string value, string &errMsg
)
{
   debug(CONFIG_ENTRY, key, value);

   Config *c = (Config *)voidPtr;

   return(
      Cfg::doString(key, value, c->mCurrentEntry->mTo, errMsg)
   );
}

int Config::entryUpdate(
   void *voidPtr, string key, string value, string &errMsg
)
{
   debug(CONFIG_ENTRY, key, value);

   Config *c = (Config *)voidPtr;

   return(
      Cfg::doBoolean(
         key, value, c->mCurrentEntry->mUpdate, errMsg
      )
   );
}

int Config::entryFrom(
   void *voidPtr, string key, string value, string &errMsg
)
{
   debug(CONFIG_ENTRY, key, value);

   Config *c = (Config *)voidPtr;

   return(Cfg::doString(key, value, c->mCurrentEntry->mFrom, errMsg));
}

int Config::entryRecurse(
   void *voidPtr, string key, string value, string &errMsg
)
{
   debug(CONFIG_ENTRY, key, value);

   Config *c = (Config *)voidPtr;

   return(
      Cfg::doBoolean(
         key, value, c->mCurrentEntry->mRecurse, errMsg
      )
   );
}

int Config::entryHidden(
   void *voidPtr, string key, string value, string &errMsg
)
{
   debug(CONFIG_ENTRY, key, value);

   Config *c = (Config *)voidPtr;

   return(
      Cfg::doBoolean(
         key, value, c->mCurrentEntry->mHidden, errMsg
      )
   );
}

int Config::entrySystem(
   void *voidPtr, string key, string value, string &errMsg
)
{
   debug(CONFIG_ENTRY, key, value);

   Config *c = (Config *)voidPtr;

   return(
      Cfg::doBoolean(
         key, value, c->mCurrentEntry->mSystem, errMsg
      )
   );
}

int Config::entryExclude(
   void *voidPtr, string key, string value, string &errMsg
)
{
   debug(CONFIG_ENTRY, key, value);

   Config *c = (Config *)voidPtr;
   string s  = "";

   if(!Cfg::doString(key, value, s, errMsg))
      return(0);

   c->mCurrentEntry->mExcludeList.push_back(s);

   return(1);
}
