/*****************************************************************************
 * C/C++ Includes
 *****************************************************************************/

#include <stdio.h>
#include <windows.h>
#include <direct.h>
#include <io.h>
#include <sys/types.h>
#include <sys/stat.h>

/*****************************************************************************
 * Local Includes
 *****************************************************************************/

#include "dir.h"
#include "util.h"
#include "regex.h"

/*****************************************************************************
 * Definition: struct DirEntry
 *****************************************************************************/

DirEntry::DirEntry() :
   mName(""),
   mFullPath(""),
   mModTime(0),
   mIsDirectory(0),
   mIsHidden(0),
   mIsSystem(0),
   mSize(0)
{
}

DirEntry::DirEntry(
   string fullPath,
   time_t modTime,
   int isDirectory,
   int isHidden,
   int isSystem,
   __int64 size
) :
   mName(""),
   mFullPath(fullPath),
   mModTime(modTime),
   mIsDirectory(isDirectory),
   mIsHidden(isHidden),
   mIsSystem(isSystem),
   mSize(size)
{
   int index = Util::lastIndexOf(fullPath, Dir::csep());

   if(index == -1)
   {
      mName = mFullPath;
   }
   else
   {
      mName = Util::substring(fullPath, index + 1);
   }

   // printf("mFullPath: [%s]\n", mFullPath.c_str());
}

void DirEntry::dump()
{
   string s = ctime(&mModTime);

   printf(
      "[%s%s%s] [%I64u] [%s] [%s]\n",
      mIsDirectory ? "D" : " ",
      mIsHidden    ? "H" : " ",
      mIsSystem    ? "S" : " ",
      mSize,
      Util::trim(s).c_str(),
      mFullPath.c_str()
   );
}

/*****************************************************************************
 * Definition: class Dir
 *****************************************************************************/

string Dir::sep()
{
   string theString = "\\";

   return(theString);
}

char Dir::csep()
{
   return('\\');
}

int Dir::hasDriveLetter(string s)
{
   int length = s.length();

   if(length < 2)
      return(0);

   char c1 = s.at(0);
   char c2 = s.at(1);

   if(c2 != ':')
      return(0);

   if(c1 >= 'a' && c1 <= 'z')
      return(1);

   if(c1 >= 'A' && c1 <= 'Z')
      return(1);

   return(0);
}

int Dir::driveExists(char letter)
{
   char buffer[3];

   sprintf(buffer, "%c:", letter);

   switch(GetDriveType(buffer))
   {
      case DRIVE_UNKNOWN    :
      case DRIVE_NO_ROOT_DIR:
      {
         return(0);
      }
      break;
   }

   return(1);
}

int Dir::isAbsolutePath(string s)
{
   if(!Dir::hasDriveLetter(s))
      return(0);

   if(s.length() < 3)
      return(0);

   if(s.at(2) != Dir::csep())
      return(0);

   return(1);
}

int Dir::cwd(string &value, string &errMsg)
{
   char buffer[4096] = { '\0' };

   if(_getcwd(buffer, sizeof(buffer) - 1) == NULL)
   {
      errMsg = "Could not get current working directory";
      return(0);
   }

   value = buffer;

   return(1);
}

int Dir::cd(string theDir, string &errMsg)
{
   if(_chdir(theDir.c_str()))
   {
      errMsg = "Could not change to directory: [" + theDir + "]";
      return(0);
   }

   return(1);
}

int Dir::getProgramPath(string &thePath, string &errMsg)
{
   char pathBuffer[4096] = { '\0' };

   DWORD value = GetModuleFileName(
      0, pathBuffer, sizeof(pathBuffer) - 1
   );

   if(!value)
   {
      errMsg = "Error getting program path";
      return(0);
   }

   thePath = pathBuffer;

   return(1);
}

int Dir::getEntry(string thePath, DirEntry &entry, string &errMsg)
{
   int isDirectory = 0;
   int isHidden    = 0;
   int isSystem    = 0;
   time_t modTime  = 0;
   __int64  size   = 0;

   if(Dir::isAbsolutePath(thePath) && thePath.length() == 3)
   {
      // No way to get actual mod time that I've found,
      // use: 'right now'.

      time(&modTime);
      isDirectory = 1;
   }
   else
   {
      struct _finddatai64_t info;

      long handle = _findfirsti64(thePath.c_str(), &info);

      if(handle == -1)
      {
         errMsg = "Could not find: [" + thePath + "]";
         return(0);
      
      }

      if((_A_SUBDIR & info.attrib) != 0)
         isDirectory = 1;

      if((_A_HIDDEN & info.attrib) != 0)
         isHidden = 1;

      if((_A_SYSTEM & info.attrib) != 0)
         isSystem = 1;

      modTime = info.time_write;
      size    = info.size;

      _findclose(handle);
   }

   DirEntry theEntry(
      thePath,
      modTime,
      isDirectory,
      isHidden,
      isSystem,
      size
   );

   entry = theEntry;

   return(1);
}

int Dir::isDir(string theDir)
{
   struct _stat s;

   if(_stat(theDir.c_str(), &s) != 0)
      return(0);

   return((s.st_mode & _S_IFDIR) != 0);
}

int Dir::isFile(string theFile)
{
   struct _stat s;

   if(_stat(theFile.c_str(), &s) != 0)
      return(0);

   return((s.st_mode & _S_IFREG) != 0);
}

string Dir::toStdForm(string path)
{
   string s  = "";
   int inSep = 0;
   int n     = path.length();

   // Turn runs of multiple directory separators into
   // single separator chars

   for(int i = 0; i < n; i ++)
   {
      char c = path.at(i);

      if(inSep)
      {
         if(c != Dir::csep())
         {
            s += c;

            inSep = 0;
         }
      }
      else
      {
         s += c;

         if(c == Dir::csep())
            inSep = 1;
      }
   }

   if((n = s.length()) > 0)
   {
      // ------------------------------------------
      // Remove terminating separator unless
      // 's' is a root directory (i.e. c:\, or \)
      // ------------------------------------------

      if(s.at(n - 1) == Dir::csep())
      {
         int removeSep = 1;

         if(n == 1)
         {
            removeSep = 0;
         }
         else if(n == 3 && Dir::isAbsolutePath(s))
         {
            removeSep = 0;
         }

         if(removeSep)
            s = Util::substring(s, 0, n - 1);
      }
   }

   return(s);
}

int Dir::ensureDir(string theDir, string &errMsg)
{
   vector<string> parts = Util::split(theDir, Dir::csep());

   int n = parts.size();

   if(n == 0)
   {
      errMsg = "Invalid directory: [" + theDir + "]";
      return(0);
   }

   string current = "";

   for(int i = 0; i < n; i ++)
   {
      string piece = parts.at(i);

      if(current != "")
         current += Dir::sep();

      current += piece;

      int handled = 0;

      if(i == 0)
      {
         if(Dir::hasDriveLetter(piece))
         {
            if(!Dir::driveExists(piece.at(0)))
            {
               errMsg = "Drive does not exist: [" + theDir + "]";
               return(0);
            }

            handled = 1;
         }
      }

      if(!handled)
      {
         // See if 'current' exists as a file ...

         if(Dir::isFile(current))
         {
            errMsg = "Not a directory: [" + current + "]";
            return(0);
         }

         if(!Dir::isDir(current))
         {
            // Try to make the directory

            if(_mkdir(current.c_str()) != 0)
            {
               errMsg = "Could not create directory: [" + current + "]";
               return(0);
            }
         }
      }
   }

   return(1);
}

int Dir::list(string thePath, DirEntryVector &theList, string &errMsg)
{
   string theString = Dir::toStdForm(thePath);

   if(!Util::endsWith(theString, Dir::sep()))
      theString += Dir::sep();

   string spec = theString + "*";

   struct _finddatai64_t info;

   long handle = _findfirsti64(spec.c_str(), &info);

   if(handle == -1)
   {
      errMsg = "No match for: [" + spec + "]";
      return(0);
   }

   // Ensure our files come first, then the directories

   DirEntryVector theDirList;
   
   int go = 1;

   while(go)
   {
      int isDirectory = 0;

      if((_A_SUBDIR & info.attrib) != 0)
         isDirectory = 1;

      int isHidden = 0;

      if((_A_HIDDEN & info.attrib) != 0)
         isHidden = 1;

      int isSystem = 0;

      if((_A_SYSTEM & info.attrib) != 0)
         isSystem = 1;

      string theFullPath = theString + info.name;

      DirEntry theEntry(
         theFullPath,
         info.time_write,
         isDirectory,
         isHidden,
         isSystem,
         info.size
      );

      if(isDirectory)
      {
         theDirList.push_back(theEntry);
      }
      else
      {
         theList.push_back(theEntry);
      }

      if(_findnexti64(handle, &info) == -1)
         go = 0;
   }

   _findclose(handle);

   for(int i = 0; i < theDirList.size(); i ++)
      theList.push_back(theDirList[i]);

   return(1);
}

int Dir::listRegex(
   string regex,
   string thePath,
   DirEntryVector &theList,
   string &errMsg
)
{
   DirEntryVector full;
  
   if(!list(thePath, full, errMsg))
      return(0);

   EntryVector instances;

   const char *pattern = regex.c_str();
   int size            = full.size();

   for(int i = 0; i < size; i ++)
   {
      DirEntry dEntry    = full.at(i);
      string matchErrMsg = "";

      int value = Regex::match(
         pattern,
         1,
         dEntry.mName.c_str(),
         instances,
         matchErrMsg
      );

      if(!value)
      {
         if(matchErrMsg != "")
         {
            errMsg = matchErrMsg;
            return(0);
         }
      }
      else
      {
         theList.push_back(dEntry);
      }
   }

   return(1);
}

#ifdef __MAIN__

static void driveExistsTest()
{
   printf(
      "Drive   Exists\n"
      "-----   ------\n"
   );

   for(char ch = 'a'; ch <= 'z'; ch ++)
   {
      printf(
         "  %c:\\      %s\n",
         ch, Dir::driveExists(ch) ? "Yes" : "No"
      );
   }
}

static void isDirTest()
{
   string list[] =
   {
      "z:x",
      "c:.",
      "c:y",
      "c::",
      "c:\\windows",
      "c:\\windows\\foo",
      "c:\\windows\\foo\\\\",
      "g:",
      "g:\\",
      "g:\\\\",
      "g:\\.",
      ".",
      ".\\.",
      "..",
      "..\\..",
      "..\\..\\..",
      "z:",
      "f:",
      "f:\\bin",
      "\\",
      "\\foo",
      "\\foo\\\\",
   };

   int n = USZ(list);

   string errMsg;

   for(int i = 0; i < n; i ++)
   {
      printf(
         "%s => isDir => %s\n",
         list[i].c_str(),
         Dir::isDir(list[i]) ? "Yes" : "No"
      );
   }
}

static void ensureDirTest()
{
   string list[] =
   {
      "c:",
      "c:.",
      "c:\\",
      "z:",
      "z:\\",
      "z:\\x",
      "z:\\x\\",
      "\\",
      "\\\\",
      ".",
      "..",
      "..\\..",
      "..\\..\\",
      "..\\..\\..",
      "..\\.\\..\\.",
      "foo",
      "test",
      "test\\"
   };

   string errMsg = "";

   int n = USZ(list);

   for(int i = 0; i < n; i ++)
   {
      if(0)
      {
         string dirStd;

         if(!Dir::ensureDir(list[i], dirStd, errMsg))
         {
            UPR(errMsg);
         }
         else
         {
            UPR(dirStd);
         }
      }
      else
      {
         printf(
            "%s => %s\n",
            list[i].c_str(), Dir::isDir(list[i]) ? "yes" : "no"
         );
      }
   }
}

int main(int argc, char **argv)
{
   string thePath = "";
   string errMsg  = "";

   if(argc < 2)
   {
      printf("Usage: %s entryName\n", argv[0]);
      exit(-1);
   }

   thePath = argv[1];

   printf(
      "[%s] => File? %s\n",
      thePath.c_str(), Dir::isFile(thePath) ? "Yes" : "No"
   );

   printf(
      "[%s] =>  Dir? %s\n",
      thePath.c_str(), Dir::isDir(thePath) ? "Yes" : "No"
   );

   DirEntry entry;

   if(!Dir::getEntry(thePath, entry, errMsg))
   {
      UPR(errMsg);
   }
   else
   {
      entry.dump();

      if(entry.mIsDirectory)
      {
         DirEntryVector v;

         if(!Dir::list(thePath, v, errMsg))
         {
            UPR(errMsg);
         }
         else
         {
            for(int i = 0, n = v.size(); i < n; i ++)
               v.at(i).dump();
         }
      }
   }

   printf("Ok\n");
   return(0);
}

#endif
