// ===========================================================================
// Revision History
// ================
// 
//    Version 1.0   - 09 Nov 2006 - Initial coding
//    Version 1.1   - 20 May 2007 - Added support for 'force' switch
//    Version 1.2   - 21 May 2007 - Handle unreadable directory when
//                                  'force' switch is enabled, also write
//                                  force error messages to html output
//    Version 1.2.1 - 12 Nov 2007 - Change "Force Error" to "Force Alert"
//    Version 1.2.2 - 20 Dec 2007 - Bad 'from' or 'to' directories
//                                  are now tolerated when 'force' is yes.
//    Version 1.2.3 - 12 Jan 2008 - Added more detailed error messages
//                                  (via strerror/errno) for read/write errors
//    Version 1.2.4 - 25 Aug 2008 - Added support for copying files larger
//                                  than 4 GB (4,294,967,296 bytes)
// ===========================================================================

/*****************************************************************************
 * C/C++ Headers
 *****************************************************************************/

#include <stdio.h>
#include <memory>
#include <sstream>
#include <errno.h>
#include <time.h>

using std::auto_ptr;
using std::ostringstream;

/*****************************************************************************
 * Local Headers
 *****************************************************************************/

#include "messagehandler.h"
#include "stringset.h"
#include "output.h"
#include "config.h"
#include "util.h"
#include "cfg.h"
#include "dir.h"

/*****************************************************************************
 * Local Defines
 *****************************************************************************/

#define PROGRAM_NAME    "zcopy"
#define VERSION         "1.2.4"
#define LOG_FILE_NAME   PROGRAM_NAME ".log"
#define CFG_NAME        PROGRAM_NAME ".cfg"
#define BUFFER_SIZE     (4096)
#define FIELD_WIDTH     (10)
#define INDENT          "   "


/*****************************************************************************
 * Local Function Prototypes
 *****************************************************************************/

static string addCommas(string s);

static int getConfig(
   const char *fileName, Config *configPtr, string &errMsg
);

static Output *getOutputToFile(int doVerbose);

static Output *getOutput(
   int doVerbose, int doConsole, int doLog
);

static int checkForDots(string path, string &errMsg);

static int checkDirectory(
   string name,
   string value,
   string errMsgPrefix,
   int checkExists,
   string &errMsg
);

static int getValues(
   Config    &config,
   int       entryIndex,
   int       &doUpdate,
   int       &doHidden,
   int       &doSystem,
   int       &doRecurse,
   string    &from,
   string    &to,
   StringSet &exclude,
   int       doForce,
   string    &errMsg
);

static int doIgnore(
   string    indent,
   DirEntry  entry,
   int       doHidden,
   int       doSystem,
   StringSet &exclude,
   Output    *out
);

static int copyFile(
   string   indent,
   string   from,
   string   to,
   __int64  size,
   Output   *out,
   int      *isOutOfDiskSpacePtr,
   string   &errMsg
);

static int copy(
   string         from,
   string         to,
   int            doUpdate,
   int            doHidden,
   int            doSystem,
   int            doRecurse,
   int            doForce,
   StringSet      &exclude,
   Output         *out,
   DirEntry       *fromEntryPtr,
   int            &totalFilesCopied,
   int            &totalFilesProcessed,
   __int64        &totalBytesCopied,
   __int64        &totalBytesProcessed,
   ForceAlertList &fList,
   string         &errMsg
);

static int checkContainment(string to, string from, string &errMsg);

static int goToWorkingDirectory(string &errMsg);

static string getBanner(string action);

static string fmtBoolean(string name, int value);
static string fmtString (string name, string value);

static void bail(string errMsg);

/*****************************************************************************
 * Local Types
 *****************************************************************************/

class Timer
{
   public:
      Timer();
      void reset();
      double getElapsed();
   private:
      time_t mBegin;
};

Timer::Timer()
{
   reset();
}

void Timer::reset()
{
   time(&mBegin);
}

double Timer::getElapsed()
{
   time_t end;

   time(&end);

   double start = mBegin;
   double stop  = end;

   double elapsed = (stop - start);

   return(elapsed);
}

/*****************************************************************************
 * Local Functions
 *****************************************************************************/

static string addCommas(string s)
{
   int n = s.length();

   if(n > 0)
   {
      string wrk = "";

      for(int i = n - 1, j = 0; i >= 0; i --, j ++)
      {
         wrk = s.at(i) + wrk;

         if(i > 0 && j % 3 == 2)
            wrk = "," + wrk;
      }

      s = wrk;
   }

   return(s);
}

static int getConfig(const char *fileName, Config *configPtr, string &errMsg)
{
   CfgLines lines;

   if(!Cfg::getLines(fileName, lines, errMsg))
      return(0);

   if(0)
   {
      lines.dump();
   }

   CfgSections sections;

   if(!Cfg::parse(lines, sections, errMsg))
      return(0);

   int n = sections.size();
   int i = 0;

   if(0)
   {
      printf("------------------------------------------------\n");
      printf("Sections - Dump\n");
      printf("------------------------------------------------\n");
      printf("\n");

      for(i = 0; i < n; i ++)
         sections.at(i).dump();
   }

   CfgItemSpec spec[] =
   {
      {
         CONFIG_OPTIONS,
         NULL,
         CFG_EXACTLY_ONE,
         NULL,
         Config::optionsBegin
      },
      {
         CONFIG_OPTIONS,
         CONFIG_TO,
         CFG_ZERO_OR_ONE,
         NULL,
         Config::optionsTo
      },
      {
         CONFIG_OPTIONS,
         CONFIG_UPDATE,
         CFG_ZERO_OR_ONE,
         CFG_NO,
         Config::optionsUpdate
      },
      {
         CONFIG_OPTIONS,
         CONFIG_HIDDEN,
         CFG_ZERO_OR_ONE,
         CFG_YES,
         Config::optionsHidden
      },
      {
         CONFIG_OPTIONS,
         CONFIG_SYSTEM,
         CFG_ZERO_OR_ONE,
         CFG_YES,
         Config::optionsSystem
      },
      {
         CONFIG_OPTIONS,
         CONFIG_VERBOSE,
         CFG_ZERO_OR_ONE,
         CFG_NO,
         Config::optionsVerbose
      },
      {
         CONFIG_OPTIONS,
         CONFIG_FORCE,
         CFG_ZERO_OR_ONE,
         CFG_NO,
         Config::optionsForce
      },
      {
         CONFIG_OPTIONS,
         CONFIG_LOG,
         CFG_ZERO_OR_ONE,
         CFG_YES,
         Config::optionsLog
      },
      {
         CONFIG_OPTIONS,
         CONFIG_CONSOLE,
         CFG_ZERO_OR_ONE,
         CFG_YES,
         Config::optionsConsole
      },
      {
         CONFIG_OPTIONS,
         CONFIG_SHOW_ERROR,
         CFG_ZERO_OR_ONE,
         CFG_YES,
         Config::optionsShowError
      },
      {
         CONFIG_OPTIONS,
         CONFIG_SHOW_SUCCESS,
         CFG_ZERO_OR_ONE,
         CFG_YES,
         Config::optionsShowSuccess
      },
      {
         CONFIG_ENTRY,
         NULL,
         CFG_ONE_OR_MORE,
         NULL,
         Config::entryBegin
      },
      {
         CONFIG_ENTRY,
         CONFIG_TO,
         CFG_ZERO_OR_ONE,
         NULL,
         Config::entryTo
      },
      {
         CONFIG_ENTRY,
         CONFIG_UPDATE,
         CFG_ZERO_OR_ONE,
         NULL,
         Config::entryUpdate
      },
      {
         CONFIG_ENTRY,
         CONFIG_FROM,
         CFG_EXACTLY_ONE,
         NULL,
         Config::entryFrom
      },
      {
         CONFIG_ENTRY,
         CONFIG_RECURSE,
         CFG_ZERO_OR_ONE,
         CFG_YES,
         Config::entryRecurse
      },
      {
         CONFIG_ENTRY,
         CONFIG_HIDDEN,
         CFG_ZERO_OR_ONE,
         NULL,
         Config::entryHidden
      },
      {
         CONFIG_ENTRY,
         CONFIG_SYSTEM,
         CFG_ZERO_OR_ONE,
         NULL,
         Config::entrySystem
      },
      {
         CONFIG_ENTRY,
         CONFIG_EXCLUDE,
         CFG_ZERO_OR_MORE,
         NULL,
         Config::entryExclude
      }
   };

   int nSpec = USZ(spec);
   void *v   = (void *)configPtr;

   if(0)
   {
      configPtr->dump("Initial");
   }

   if(!Cfg::applyDefaults(v, spec, nSpec, errMsg))
      return(0);

   if(0)
   {
      configPtr->dump("Defaults Applied");
   }

   if(!Cfg::processConfig(v, 0, sections, spec, nSpec, errMsg))
      return(0);

   if(0)
   {
      configPtr->dump("Final");
   }

   return(1);
}

static Output *getOutputToFile(int doVerbose)
{
   Output *out = new OutputToFile(doVerbose, LOG_FILE_NAME);

   return(out);
}

static Output *getOutput(int doVerbose, int doConsole, int doLog)
{
   Output *oc = NULL;

   if(doConsole)
      oc = new OutputToConsole(doVerbose);

   if(doLog)
   {
      Output *of = getOutputToFile(doVerbose);

      return(doConsole ? new OutputTee(oc, of) : of);
   }

   return(oc);
}

static int checkForDots(string path, string &errMsg)
{
   vector<string> parts = Util::split(path, Dir::csep());

   for(int n = parts.size(), i = 0; i < n; i ++)
   {
      if(parts[i] == "." || parts[i] == "..")
      {
         errMsg =
            "(" + path + "), cannot contain '.' or '..'";

         return(0);
      }
   }

   return(1);
}

static int checkDirectory(
   string name,
   string value,
   string errMsgPrefix,
   int checkExists,
   string &errMsg
)
{
   if(!Dir::isAbsolutePath(value))
   {
      errMsg =
         errMsgPrefix +
         ", value: '" + name + "' (" + value + "), not an absolute path";

      return(0);
   }

   if(checkExists)
   {
      if(!Dir::isDir(value))
      {
         errMsg =
            errMsgPrefix +
            ", value: '" + name + "' (" + value + "), not a directory";

         return(0);
      }
   }

   if(!checkForDots(value, errMsg))
   {
      errMsg = errMsgPrefix + ", value: '" + name + "' " + errMsg;
      return(0);
   }

   return(1);
}

static int getValues(
   Config    &config,
   int       entryIndex,
   int       &doUpdate,
   int       &doHidden,
   int       &doSystem,
   int       &doRecurse,
   string    &from,
   string    &to,
   StringSet &exclude,
   int       doForce,
   string    &errMsg
)
{
   Config::Entry theEntry = config.mEntries.at(entryIndex);

   // ---------------------------------------------------
   // For the: update, hidden and system settings, we
   // default to our 'options' settings but use our
   // entry setting if present
   // ---------------------------------------------------

   struct ValuesEntry
   {
      string optionsString;
      string entryString;
      int    &dest;

      ValuesEntry(string oString, string eString, int &ref) :
         optionsString(oString),
         entryString(eString),
         dest(ref)
      {
      }
   };

   ValuesEntry entries[] =
   {
      ValuesEntry(config.mOptions.mUpdate, theEntry.mUpdate, doUpdate),
      ValuesEntry(config.mOptions.mHidden, theEntry.mHidden, doHidden),
      ValuesEntry(config.mOptions.mSystem, theEntry.mSystem, doSystem)
   };

   int size = USZ(entries);
   int i    = 0;

   for(i = 0; i < size; i ++)
   {
      string s = entries[i].optionsString;

      if(entries[i].entryString != "")
         s = entries[i].entryString;

      entries[i].dest = Cfg::isTrue(s);
   }

   doRecurse = Cfg::isTrue(theEntry.mRecurse);

   string entryErrMsg =
      "entry[" + Util::toString(1 + entryIndex) + "]";

   // -----------------------------------
   // Now, handle 'to' and 'from' ...
   // -----------------------------------

   if(config.mOptions.mTo == "" && theEntry.mTo == "")
   {
      errMsg =
         entryErrMsg + ": 'to' not set in entry or options";

      return(0);
   }

   from = theEntry.mFrom;
   to   = (theEntry.mTo == "" ? config.mOptions.mTo : theEntry.mTo);

   // ------------------------------------------------
   // The values for 'from' and 'to' must:
   //    - be absolute paths: (i.e. g:\, c:\stuff)
   //    - be existing directories
   //    - not contain any '.' or '..' directories
   // ------------------------------------------------

   from = Dir::toStdForm(from);
   to   = Dir::toStdForm(to);

   int fromCheckExists = 1;
   int toCheckExists   = 0;

   if(doForce)
   {
      fromCheckExists = 0;
      toCheckExists   = 0;
   }

   if(!checkDirectory("from", from, entryErrMsg, fromCheckExists, errMsg))
      return(0);

   if(!checkDirectory("to", to, entryErrMsg, toCheckExists, errMsg))
      return(0);

   to = Dir::toStdForm(to);

   vector<string> excludeList = theEntry.mExcludeList;
   size                       = excludeList.size();
   
   for(i = 0; i < size; i ++)
   {
      string xString = excludeList.at(i);

      if(!Dir::isAbsolutePath(xString))
      {
         errMsg =
            entryErrMsg +
            ", value: 'exclude' (" + xString + "), not an absolute path";
   
         return(0);
      }

      if(!checkForDots(xString, errMsg))
      {
         errMsg = entryErrMsg + ", value: 'exclude' " + errMsg;
         return(0);
      }

      exclude.insert(Dir::toStdForm(xString));
   }

   return(1);
}

static int doIgnore(
   string    indent,
   DirEntry  entry,
   int       doHidden,
   int       doSystem,
   StringSet &exclude,
   Output    *out
)
{
   string which = (entry.mIsDirectory ? "directory" : "file");

   if(exclude.contains(entry.mFullPath))
   {
      out->writeVerbose(
         indent + "Excluding " + which + " [" + entry.mFullPath + "]"
      );

      return(1);
   }

   if(!doHidden && entry.mIsHidden)
   {
      out->writeVerbose(
         indent + "Excluding hidden " + which + " [" + entry.mFullPath + "]"
      );

      return(1);
   }

   if(!doSystem && entry.mIsSystem)
   {
      out->writeVerbose(
         indent + "Excluding system " + which + " [" + entry.mFullPath + "]"
      );

      return(1);
   }

   return(0);
}

// -------------------------------------------------------------------
// JFD - Stuff we still need to handle ...
//
//    - Ensure we behave sanely when we run out of disk space
//      on our dest drive.
// -------------------------------------------------------------------

static int copyFile(
   string   indent,
   string   from,
   string   to,
   __int64  size,
   Output   *out,
   int      *isOutOfDiskSpacePtr,
   string   &errMsg
)
{
   FILE *fIn                         = NULL;
   FILE *fOut                        = NULL;
   unsigned char buffer[BUFFER_SIZE] = { 0 };
   int blocks                        = size / BUFFER_SIZE;
   int extraBytes                    = size % BUFFER_SIZE;

   ostringstream errStream;

   *isOutOfDiskSpacePtr = 0;

   if(extraBytes)
      blocks++;

   out->writeVerbose(
      indent + "Copying: [" + from + "] to: [" + to + "]"
   );

   if((fIn = fopen(from.c_str(), "rb")) == NULL)
   {
      errStream <<
         "Could not open: [" <<
         from <<
         "] for reading " <<
         "(" <<
         strerror(errno) <<
         ")";

      errMsg = errStream.str();

      return(0);
   }

   if((fOut = fopen(to.c_str(), "wb")) == NULL)
   {
      fclose(fIn);

      errStream <<
         "Could not open: [" <<
         to <<
         "] for writing " <<
         "(" <<
         strerror(errno) <<
         ")";

      errMsg = errStream.str();

      return(0);
   }

   int returnValue = 1;
   __int64 offset  = 0;

   for(int i = 0; returnValue && i < blocks; i ++)
   {
      int n = (i + 1 == blocks && extraBytes ? extraBytes : BUFFER_SIZE);

      if(fread(buffer, 1, n, fIn) != n)
      {
         errMsg =
            "Error reading from: [" + from + "], " +
            "offset: [" + Util::toString(offset) + "]";

         returnValue = 0;
      }
      else if(fwrite(buffer, 1, n, fOut) != n)
      {
         errMsg =
            "Error writing to: [" + to + "], " +
            "offset: [" + Util::toString(offset) + "]";

         // -------------------------------------------------
         // If we've run out of disk space, strerror
         // will supply us with 'No space left on device'
         // which will make our error message more
         // comprehensible to the user
         // -------------------------------------------------

         if(errno != 0)
         {
            ostringstream o;

            o << errMsg + " [" << strerror(errno) << "] " <<
               "(errno: " << errno << ")";

            errMsg = o.str();

            if(errno == ENOSPC)
            {
               *isOutOfDiskSpacePtr = 1;
            }
         }

         returnValue = 0;
      }

      offset += n;
   }

   fclose(fIn);
   fclose(fOut);

   return(returnValue);
}

static void handleForceErrMsg(
   ForceAlertList &fList, Output *out, const string &errMsg
)
{
   string theMessage =
      "-- (Force Alert) -- " + errMsg;

   fList.push_back(theMessage);
   out->write(theMessage);
}

static int copy(
   string         indent,
   string         from,
   string         to,
   int            doUpdate,
   int            doHidden,
   int            doSystem,
   int            doRecurse,
   int            doForce,
   StringSet      &exclude,
   Output         *out,
   DirEntry       *fromEntryPtr,
   int            &totalFilesCopied,
   int            &totalFilesProcessed,
   __int64        &totalBytesCopied,
   __int64        &totalBytesProcessed,
   ForceAlertList &fList,
   string    &errMsg
)
{
   string line =
      "-------------------------------------------------------------";

   out->writeVerbose(indent + line);
   out->writeVerbose(indent + fmtString("from", from));
   out->writeVerbose(indent + fmtString("to",   to));
   out->writeVerbose(indent + line);

   indent += INDENT;

   // -------------------------------------------------
   // If we're the top-level call to 'copy' we'll
   // need to get our Entry info about 'from'.
   // For non-top-level calls, we'll be passed
   // a pointer to from's Entry struct.
   // -------------------------------------------------

   DirEntry fromEntry;

   if(fromEntryPtr == NULL)
   {
      if(!Dir::getEntry(from, fromEntry, errMsg))
      {
         errMsg = "Could not read 'from' directory: [" + from + "]";

         if(!doForce)
            return(0);

         handleForceErrMsg(fList, out, errMsg);
         return(1);
      }
   }
   else
   {
      fromEntry = *fromEntryPtr;
   }

   if(doIgnore(indent, fromEntry, doHidden, doSystem, exclude, out))
      return(1);

   // ----------------------------------------------------
   // Ok, now we know that 'from' has not been excluded
   // nor weeded out by doHidden/doSystem.  We now
   // need to ensure that 'to' exists as a directory.
   // ----------------------------------------------------

   if(!Dir::ensureDir(to, errMsg))
   {
      if(!doForce)
         return(0);

      handleForceErrMsg(fList, out, errMsg);
      return(1);
   }

   DirEntryVector theList;

   if(!Dir::list(from, theList, errMsg))
   {
      // We could not get a list of from's files and
      // subdirectories (not even '.' and '..').
      // Most likely this means we were prohibited from
      // even reading the directory as happens when we
      // copy the 'c:\system volume information folder'.

      
      if(!doForce)
         return(0);

      handleForceErrMsg(fList, out, errMsg);
      return(1);
   }

   for(int n = theList.size(), i = 0; i < n; i ++)
   {
      DirEntry entry  = theList.at(i);
      string destPath = to + Dir::sep() + entry.mName;

      if(entry.mIsDirectory)
      {
         if(entry.mName != "." && entry.mName != "..")
         {
            if(!doRecurse)
            {
               out->writeVerbose(
                  indent + "Not recursing into: [" + entry.mFullPath + "]"
               );
            }
            else
            {
               int status = copy(
                  indent,
                  entry.mFullPath,
                  destPath,
                  doUpdate,
                  doHidden,
                  doSystem,
                  doRecurse,
                  doForce,
                  exclude,
                  out,
                  &entry,
                  totalFilesCopied,
                  totalFilesProcessed,
                  totalBytesCopied,
                  totalBytesProcessed,
                  fList,
                  errMsg
               );

               if(!status)
                  return(0);
            }
         }
      }
      else
      {
         if(!doIgnore(indent, entry, doHidden, doSystem, exclude, out))
         {
            int doCopyFile = 1;

            // Update only?

            if(doUpdate)
            {
               // If the dest version of this file exists,
               // is the same size and is no older than the
               // src version of the file, we don't copy it

               DirEntry destEntry;

               if(Dir::getEntry(destPath, destEntry, errMsg))
               {
                  if(entry.mSize == destEntry.mSize)
                  {
                     if(entry.mModTime <= destEntry.mModTime)
                     {
                        doCopyFile = 0;

                        out->writeVerbose(
                           indent + 
                           "File: [" + entry.mFullPath + "], no need to copy"
                        );
                     }
                  }
               }
            }

            if(doCopyFile)
            {
               int isOutOfDiskSpace = 0;

               int status = copyFile(
                  indent,
                  entry.mFullPath,
                  destPath,
                  entry.mSize,
                  out,
                  &isOutOfDiskSpace,
                  errMsg
               );

               if(status)
               {
                  totalFilesCopied++;
                  totalBytesCopied += entry.mSize;
               }
               else
               {
                  if(isOutOfDiskSpace)
                  {
                     // Try to remove the file, no point in leaving
                     // it around ...


                     if(unlink(destPath.c_str()) != 0)
                     {
                        errMsg +=
                           ", Could not remove file: [" + destPath + "]";
                     }

                     return(0);
                  }

                  // If we're not out of disk space, we'll keep going
                  // if force is set to yes


                  if(doForce)
                  {
                     handleForceErrMsg(fList, out, errMsg);
                  }
                  else
                  {
                     return(0);
                  }
               }
            }
         }

         totalFilesProcessed++;
         totalBytesProcessed += entry.mSize;
      }
   }

   return(1);
}

static int checkContainment(string to, string from, string &errMsg)
{
   // --------------------------------------
   // We disallow backing up a directory
   // to a subdirectory of itself,
   // for example:
   //
   //    from -> c:\
   //    to   -> c:\backup
   //
   // Note that since 'c:\' includes
   // 'c:\backup' anything we write
   // to 'c:\backup' means additional
   // data to be backed up from 'c:\'.
   //
   // To avoid the prospect of an
   // infinite loop we issue an error
   // when we encounter the above.
   //
   // Note that the reverse of the
   // above scenario:
   //
   //    from -> c:\backup
   //    to   -> c:\
   //
   // causes no trouble and is allowed
   // --------------------------------------

   string fromLower = Util::toLowerCase(from);
   string toLower   = Util::toLowerCase(to);

   if(!Util::startsWith(toLower, fromLower))
      return(1);

   int fromContainsTo = 0;

   // Our 'from' directory has already been
   // validated as an absolute path.  If
   // it's length is 3 then it's a drive
   // root as in 'c:\' or 'g:\' and it
   // definitely contains 'to'.

   if(fromLower.length() == 3)
   {
      fromContainsTo = 1;
   }
   else
   {
      // Watch out for false positives, as in:
      //
      //    from -> c:\backup
      //    to   -> c:\backups
      //
      // actual matches are like:
      //
      //    from -> c:\backup
      //    to   -> c:\backup\stuff
      //
      // we look at what's left of 'to'
      // after removing 'from' from the
      // beginning of it to distinguish
      // the two scenarios above.

      string rhs = Util::substring(toLower, fromLower.length());

      if(rhs == "" || rhs.at(0) == Dir::csep())
      {
         fromContainsTo = 1;
      }
   }

   if(fromContainsTo)
   {
      errMsg =
         "From directory: [" + from + "] " +
         "contains to directory: [" + to + "]";

      return(0);
   }

   return(1);
}

static int goToWorkingDirectory(string &errMsg)
{
   // ---------------------------------------------
   // We always want to run in the directory where
   // our exe resides so work out the path to
   // ourself and cd there.
   // ---------------------------------------------

   string programPath = "";

   if(!Dir::getProgramPath(programPath, errMsg))
      return(0);

   if(0)
   {
      printf("Program Path: [%s]\n", programPath.c_str());
   }

   string programDir = Util::substring(
      programPath,
      0,
      1 + Util::lastIndexOf(programPath, Dir::csep())
   );

   programDir = Dir::toStdForm(programDir);

   if(0)
   {
      printf("Program Dir: [%s]\n", programDir.c_str());
   }

   if(!Dir::cd(programDir, errMsg))
      return(0);

   return(1);
}

static string getBanner(string action)
{
   string bar =
      "=====================================================================";

   string greeting = PROGRAM_NAME;
   greeting       += " (c) - Version: [";
   greeting       += VERSION;
   greeting       += "]";

   return(
      bar + "\n" +
      greeting + " - " + action + ": [" + Util::getTimeStamp() + "]\n" +
      bar
   );
}

static string fmtBoolean(string name, int value)
{
   string s =
      Util::rpad(name, FIELD_WIDTH) + ": " + (value ? "[Yes]" : "[No]");

   return(s);
}

static string fmtString(string name, string value)
{
   string s =
      Util::rpad(name, FIELD_WIDTH) + ": [" + value + "]";

   return(s);
}

static void bail(string errMsg)
{
   Output *out = getOutputToFile(1);

   errMsg =
      "-----------------------------------------------------------\n" 
      " Fatal Error: " + errMsg + "\n"
      "-----------------------------------------------------------";

   out->write(errMsg);
   printf("%s\n", errMsg.c_str());

   delete out;

   exit(-1);
}

/*****************************************************************************
 * Main
 *****************************************************************************/

int main(int argc, char **argv)
{
   Timer theTimer;

   string errMsg = "";

   // ------------------------------------------------
   // We want to run in the directory where our exe
   // lives.  Try to arrange this and exit if
   // we cannot.
   // ------------------------------------------------

   if(!goToWorkingDirectory(errMsg))
      bail(errMsg);

   const char *fileName = CFG_NAME;

   if(argc > 1)
      fileName = argv[1];

   // --------------------------------------------
   // Acquire config settings and create our
   // Output object.
   // --------------------------------------------

   Config config;

   if(!getConfig(fileName, &config, errMsg))
      bail(errMsg);

   // Prepare our output/logging object

   int doVerbose = Cfg::isTrue(config.mOptions.mVerbose);
   int doForce   = Cfg::isTrue(config.mOptions.mForce);
   int doConsole = Cfg::isTrue(config.mOptions.mConsole);
   int doLog     = Cfg::isTrue(config.mOptions.mLog);

   if(!doConsole && !doLog)
      bail("Cannot disable both console and log output");

   Output *out = getOutput(doVerbose, doConsole, doLog);
   auto_ptr<Output> outReaper(out);

   // ------------------------------------------
   // Starting here, any errors we encounter
   // will be communicated to the user
   // via out Output object
   // ------------------------------------------

   out->write(getBanner("Starting") + "\n");

   // ------------------------------------------
   // Dump config settings (no output unless
   // we're in verbose mode)
   // ------------------------------------------

   out->writeVerbose(config.toString("Settings", 0));

   // --------------------------------------------
   // Perpare our MessageHandler which governs
   // all direct communication with the user
   // --------------------------------------------

   MessageHandler *handler = MessageHandler::get(
      Cfg::isTrue(config.mOptions.mShowError),
      Cfg::isTrue(config.mOptions.mShowSuccess),
      errMsg
   );

   if(!handler)
   {
      out->write("MessageHandler error - " + errMsg);
      exit(-1);
   }

   auto_ptr<MessageHandler> messageHandlerReaper(handler);

   // ----------------------------------------------
   // From here on out, in addition to using our
   // Output object, we'll also create an html file
   // on success/error that we'll display to the
   // user.  The html display (on success/error)
   // is governed by the show_success/show_error
   // cfg file settings.
   // ----------------------------------------------

   out->writeVerbose(
      "------------------------------------------------\n"
      "Processing Entries\n"
      "------------------------------------------------\n"
   );

   // ---------------------------------
   // We're ready for action ...
   // ---------------------------------

   int     totalFilesProcessed = 0;
   int     totalFilesCopied    = 0;
   __int64 totalBytesProcessed = 0;
   __int64 totalBytesCopied    = 0;

   int nEntries = config.mEntries.size();
   int i        = 0;

   ForceAlertList fList;

   for(i = 0; i < nEntries; i ++)
   {
      int doUpdate  = 0;
      int doHidden  = 0;
      int doSystem  = 0;
      int doRecurse = 0;
      string from   = "";
      string to     = "";

      StringSet exclude(1);

      out->writeVerbose(
         "[entry: " +
         Util::toString(i + 1) +
         " / " +
         Util::toString(nEntries) +
         "]"
      );

      int status = getValues(
         config,
         i,
         doUpdate,
         doHidden,
         doSystem,
         doRecurse,
         from,
         to,
         exclude,
         doForce,
         errMsg
      );

      if(!status)
         break;

      if(!checkContainment(to, from, errMsg))
         break;

      out->writeVerbose(INDENT + fmtBoolean("doUpdate",  doUpdate));
      out->writeVerbose(INDENT + fmtBoolean("doHidden",  doHidden));
      out->writeVerbose(INDENT + fmtBoolean("doSystem",  doSystem));
      out->writeVerbose(INDENT + fmtBoolean("doRecurse", doRecurse));

      if(exclude.count() > 0)
      {
         vector<string> list = exclude.getList();

         for(int nList = list.size(), j = 0; j < nList; j ++)
         {
            out->writeVerbose(INDENT + fmtString("exclude", list[j]));
         }
      }

      out->writeVerbose("");

      // -------------------------------------------------------
      // Construct the destination directory name, if we had ...
      //
      //    from = c:\windows\program files
      //    to   = g:\backup
      //
      // then our destination directory would be:
      //
      //    g:\backup\c\windows\program\files
      // -------------------------------------------------------

      if(!Util::endsWith(to, Dir::sep()))
         to += Dir::sep();

      to = Dir::toStdForm(to + from.at(0) + Util::substring(from, 2));

      status = copy(
         INDENT,
         from,
         to,
         doUpdate,
         doHidden,
         doSystem,
         doRecurse,
         doForce,
         exclude,
         out,
         0,
         totalFilesCopied,
         totalFilesProcessed,
         totalBytesCopied,
         totalBytesProcessed,
         fList,
         errMsg
      );

      if(!status)
         break;
   }

   if(i < nEntries)
   {
      out->write(errMsg);
      handler->exitError(fList, errMsg);

      exit(-1);
   }

   out->writeVerbose(
      "\n"
      "------------------------------------------------\n"
      "Processing Complete\n"
      "------------------------------------------------\n"
   );

   string tfcString = addCommas(Util::toString(totalFilesCopied));
   string tfpString = addCommas(Util::toString(totalFilesProcessed));

   string tbcString = addCommas(Util::toString(totalBytesCopied));
   string tbpString = addCommas(Util::toString(totalBytesProcessed));

   string pMsg =
      "Processed: [" + tfpString + "] file(s), " +
      "[" + tbpString + "] bytes";

   string cMsg =
      "Copied: [" + tfcString + "] file(s), " +
      "[" + tbcString + "] bytes";

   string eMsg =
      "Elapsed: [" +
      Util::toString("%.0f", theTimer.getElapsed()) +
      "] second(s)";

   out->writeVerbose("");

   out->write(pMsg + "\n" + cMsg + "\n" + eMsg + "\n");
   out->write(getBanner("Ending"));
   out->write("");

   handler->exitSuccess(
      fList, pMsg + "<br>\n" + cMsg + "<br>\n" + eMsg
   );

   return(0);
}
